# Copyright 2014-2017 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""User-contributed and possibly not fully supported modules within ODL."""

__all__ = ()

# Don't add any imports here since we don't want anything in this package to
# be imported by default. That way we make sure that sub-optimal stuff here
# doesn't break or bog down the import of ODL as a whole.
